use specs::{saveload::MarkedBuilder, Builder, Entity, World, WorldExt, WriteStorage};

use glc::vector::Vector2f;

use crate::{
    components::{Asteroid, AsteroidType, MeetingPoint, Obstacle, PlayerOwned, Position, Shape},
    misc::{Persistence, WorldPersistence},
};

use super::Error;

#[derive(Default, Clone)]
pub struct AsteroidBuilder {
    type_: Option<AsteroidType>,
    position: Option<Vector2f>,
    size: Option<f32>,
    orbit: Option<(f32, f32)>,
    owner: Option<Entity>,
}

impl AsteroidBuilder {
    pub fn build(&self, world: &mut World) -> Result<Entity, Error> {
        let type_ = self.type_.ok_or(Error::MissingValue("type"))?;
        let position = self.position.ok_or(Error::MissingValue("position"))?;
        let size = self.size.ok_or(Error::MissingValue("size"))?;
        let (orbit_min, orbit_max) = self.orbit.ok_or(Error::MissingValue("orbit"))?;

        let position = Position::new(position);
        let shape = Shape::Circle(size);
        let meeting_point = MeetingPoint::new(orbit_min, orbit_max);
        let obstacle = Obstacle::default();
        let asteroid = Asteroid::new(type_);

        let entity = world
            .create_entity()
            .marked::<<WorldPersistence as Persistence>::Marker>()
            .with(position)
            .with(shape)
            .with(meeting_point)
            .with(obstacle)
            .with(asteroid)
            .build();

        if let Some(owner) = self.owner {
            let mut player_owned = world.system_data::<WriteStorage<PlayerOwned>>();

            player_owned.insert(entity, PlayerOwned::new(owner))?;
        }

        Ok(entity)
    }

    pub fn type_(mut self, value: AsteroidType) -> Self {
        self.type_ = Some(value);

        self
    }

    pub fn position<V: Into<Vector2f>>(mut self, value: V) -> Self {
        self.position = Some(value.into());

        self
    }

    pub fn size(mut self, value: f32) -> Self {
        self.size = Some(value);

        self
    }

    pub fn orbit(mut self, min: f32, max: f32) -> Self {
        self.orbit = Some((min, max));

        self
    }

    pub fn owner(mut self, player: Entity) -> Self {
        self.owner = Some(player);

        self
    }
}
